package com.example.guitar_music_app.login

import androidx.lifecycle.MutableLiveData
import com.example.guitar_music_app.general.*
import com.example.guitar_music_app.general.ANTENNA_EMPTY
import com.example.guitar_music_app.general.ANTENNA_FULL
import com.example.guitar_music_app.general.ANTENNA_LOOP
import com.example.guitar_music_app.general.LOADING
import com.example.guitar_music_app.general.LOGIN_ERROR
import com.example.guitar_music_app.general.RC_SIGN_IN
import com.example.guitar_music_app.general.SIGNED_IN
import com.example.guitar_music_app.general.SIGNED_OUT
import com.example.guitar_music_app.general.SIGN_IN
import com.example.guitar_music_app.general.SIGN_OUT
import com.example.guitar_music_app.general.BaseViewModel
import kotlinx.coroutines.launch
import kotlin.coroutines.CoroutineContext

class LoginViewModel(
    val userRepo: UserRepository,
    uiContext: CoroutineContext
) : BaseViewModel<LoginEvent<UserLoginResult>>(uiContext) {

    //MutableLiveObject musi obsahovat vsechny objekty z View
    // Doopravdovy data model je privatni, aby nebyl nechtene menen
    private val userState = MutableLiveData<User>()

    //Control Logic
    //Internal = Viditelne jen v modulu, lepsi encapsulation
    //Unit => Jako Void v jave, trigger eventu ale nezajima nas value
    //Neprimy callback na view
    internal val authAttempt = MutableLiveData<Unit>()
    internal val startAnimation = MutableLiveData<Unit>()

    //UI Binding
    internal val signInStatusText = MutableLiveData<String>()
    internal val authButtonText = MutableLiveData<String>()
    internal val antennaDrawable = MutableLiveData<String>()

    internal val signedIn = MutableLiveData(false)

    private fun showSignedOutState() {
        signInStatusText.value = SIGNED_OUT
        authButtonText.value = SIGN_IN
        antennaDrawable.value = ANTENNA_EMPTY
    }

    private fun showErrorState() {
        signInStatusText.value = LOGIN_ERROR
        authButtonText.value = SIGN_IN
        antennaDrawable.value = ANTENNA_EMPTY
    }

    private fun showSignedInState() {
        signInStatusText.value = SIGNED_IN
        authButtonText.value = SIGN_OUT
        antennaDrawable.value = ANTENNA_FULL
    }

    private fun showLoadingState() {
        signInStatusText.value = LOADING
        antennaDrawable.value = ANTENNA_LOOP
        startAnimation.value = Unit
    }



    override fun handleEvent(event: LoginEvent<UserLoginResult>) {
        //Prvni trigruje loading screen
        showLoadingState()
        //When je switch
        when (event) {
            is LoginEvent.OnStart -> {getUser()
            }
            is LoginEvent.OnAuthButtonClick -> onAuthButtonClick()
            is LoginEvent.OnGoogleSignInResult -> onSignInResult(event.result)
        }
    }

    fun signOutUser() = launch {
        val result = userRepo.signOutCurrentUser()
        when (result) {
            is GeneralResult.Value -> {
                userState.value = null
                signedIn.value = false
                showSignedOutState()
            }
            is GeneralResult.Error -> showErrorState()
        }
    }

    private fun getUser() = launch {
        val result = userRepo.getCurrentUser()
        when (result) {
            is GeneralResult.Value -> {
                userState.value = result.value
                if (result.value == null) showSignedOutState()
                else showSignedInState()
            }
            is GeneralResult.Error -> showErrorState()
        }
    }




    //Firebase a google singIn
    private fun onSignInResult(result: UserLoginResult) = launch {
        if (result.requestCode == RC_SIGN_IN && result.userToken != null) {

            val createGoogleUserResult = userRepo.signInGoogleUser(
                result.userToken
            )

            //Result.Value je not success
            if (createGoogleUserResult is GeneralResult.Value) getUser()
            else showErrorState()

            signedIn.value = true
        } else {
            showErrorState()
        }
    }

//    Jestlize je user nula, View zacne authAttempt, jinak je uzivatel odhlasen

    private fun onAuthButtonClick() {
        //Triggruje authAttempt observable v LoginView, ktery zavola startSignInFlow funkci a ta pripravi aplikaci na prihlaseni
        if (userState.value == null) {
            authAttempt.value = Unit
        }
        else {
            signOutUser()
        }
    }



}